package org.kantega.jexmec.guice;

/*
 * Copyright 2011 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import com.google.inject.*;
import com.google.inject.name.Names;
import org.kantega.jexmec.PluginLoader;
import org.kantega.jexmec.ServiceKey;
import org.kantega.jexmec.ServiceLocator;
import org.kantega.jexmec.ctor.ClassLocator;
import org.kantega.jexmec.services.Services;

import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

/**
 * Implementation of {@link PluginLoader} based on <a href="http://code.google.com/p/google-guice/">Google Guice</a>.
 * Plugin classes are located by reading the file {@code META-INF/services/&lt;pluginClassName&gt;/guice.txt} from the
 * ClassLoader. This file may:
 * <ul>
 * <li>list plugin classes directly</li>
 * <li>list Guice {@link Module} classes which bind plugin implementations and any plugin-internal classes</li>
 * <li>use any combination of the above</li>
 * </ul>
 *
 * The plugins are then configured in a Guice Module together with all services provided
 * by the ServiceLocator. Finally, an Injector is created with the module and a list of plugins is returned using
 * Guice's {@link com.google.inject.Injector#findBindingsByType(com.google.inject.TypeLiteral)} method.
 */
public class GuicePluginLoader<P>  implements PluginLoader<P> {

    private final Class servicesInterface;

    /**
     * Creates a GuicePluginLoader locating classes using a custom class locator. The plugin loader
     * will inject the services available in the given <code>servicesInterface</code>.
     * @param servicesInterface the services interface to inject services for.
     */
    public GuicePluginLoader(Class servicesInterface) {
        Services.validateServiceInterface(servicesInterface);
        this.servicesInterface = servicesInterface;

    }

    /**
     * Locates, loads and returns any Guice plugins found in the provided ClassLoader
     * @param classLoader the ClassLoader used to locate Guice
     * @param serviceLocator the {@link org.kantega.jexmec.ServiceLocator} where services can be fetched.
     * @return the list of Guice Plugins
     */
    public List<P> loadPlugins(final Class<P> pluginClass, ClassLoader classLoader, final ServiceLocator serviceLocator) {


        ClassLocator<Class<P>> classLocator = createPluginClassLocator(pluginClass);
        final List<Class<P>> pluginClasses = classLocator.locateClasses(classLoader);

        List<Module> modules = new ArrayList<Module>();

        // Bind plugin classes
        modules.add(new Module() {
            public void configure(Binder binder) {
                for(Class<P> pc : pluginClasses) {
                    if(pluginClass.isAssignableFrom(pc)) {
                        binder.bind(pluginClass).annotatedWith(Names.named(pc.getName())).to(pc);
                    }
                }
            }
        });
        // Bind plugin Modules
        modules.add(new Module() {
            public void configure(Binder binder) {
                for(Class<P> pc : pluginClasses) {
                    if(Module.class.isAssignableFrom(pc)) {
                        try {
                            binder.install((Module) pc.newInstance());
                        } catch (InstantiationException e) {
                            throw new RuntimeException(e);
                        } catch (IllegalAccessException e) {
                            throw new RuntimeException(e);
                        }
                    }

                }
            }
        });

        // Bind services
        modules.add(new Module() {
            @SuppressWarnings("unchecked")
            public void configure(Binder binder) {


                for(final Method method : servicesInterface.getMethods()) {

                    Set<ServiceKey> serviceKeys = Services.getServiceKeysForMethod(method);

                    for(final ServiceKey serviceKey : serviceKeys) {
                        Provider serviceProvider = new Provider() {
                            public Object get() {
                                return serviceLocator.lookupService(serviceKey);
                            }
                        };

                        // Bind with type literal for parametrized return types 
                        if(method.getGenericReturnType() instanceof ParameterizedType) {

                            final ParameterizedType parameterizedReturnType = ParameterizedType.class.cast(method.getGenericReturnType());

                            final TypeLiteral<?> serviceTypeLiteral = TypeLiteral.get(new ParameterizedType() {
                                public Type[] getActualTypeArguments() {
                                    return parameterizedReturnType.getActualTypeArguments();
                                }

                                public Type getRawType() {
                                    return serviceKey.getServiceType();
                                }

                                public Type getOwnerType() {
                                    return null;
                                }
                            });


                            binder.bind(serviceTypeLiteral).toProvider(serviceProvider).in(Singleton.class);

                        } else {
                            // Bind the class serviceKey directly
                            binder.bind(getGuiceKey(serviceKey)).toProvider(serviceProvider).in(Singleton.class);
                        }

                    }
                }


            }
        });

        final Injector injector = Guice.createInjector(modules);

        List<P> plugins = new ArrayList<P>();

        for(Binding<P> binding : injector.findBindingsByType(TypeLiteral.get(pluginClass))) {

            plugins.add(binding.getProvider().get());
        }
        return plugins;
    }

    /**
     * Creates a DefaultPluginClassLocator. The default implementation creates a DefaultPluginClassLocator
     * looking for a file "guice.txt" in the plugin path <code>META-INF/services/PluginClass/</code>. Subclasses
     * can override this method to look for plugins and Guice modules in custom locations.
     * @param pluginClass the plugin class
     * @return the DefaultPluginClassLocator
     */
    protected ClassLocator<Class<P>> createPluginClassLocator(Class<P> pluginClass) {
        return new ClassLocator<Class<P>>(pluginClass, "guice.txt");
    }

    /**
     * Returns a Guice {@link Key} for the Jexmec {@link ServiceKey}. If the Jexmec service key is named, return a Guice
     * {@link com.google.inject.name.Named} key.
     * @param serviceKey the Jexmec service key
     * @return the Guice key
     */
    private <T> Key<T> getGuiceKey(ServiceKey<T> serviceKey) {
        return serviceKey.getServiceName() != null
                ? Key.get(serviceKey.getServiceType(), Names.named(serviceKey.getServiceName()))
                : Key.get(serviceKey.getServiceType());
    }

    public void unloadPlugins(ClassLoader classLoader) {
    }

    public void start() {

    }

    public void stop() {
        
    }
}
