package org.kantega.jexmec.guice;

/*
 * Copyright 2011 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import com.google.inject.Injector;
import com.google.inject.Key;
import com.google.inject.name.Names;
import org.kantega.jexmec.ServiceKey;
import org.kantega.jexmec.ServiceLocator;
import org.kantega.jexmec.services.Services;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * {@link ServiceLocator} implementation where services are looked up in a Guice module.
 */
public class GuiceServiceLocator implements ServiceLocator {
    private final Map<ServiceKey, Object> services;
    private final Set<ServiceKey> serviceKeys;

    /**
     * Creates a ServiceLocator that will look up service classes in the provided Guice module.
     *
     * @param injector        the injector to look in
     * @param servicesClass the service types that will be looked up
     */
    public GuiceServiceLocator(Injector injector, Class servicesClass) {
        this(injector, Services.getServiceKeysFromInterface(servicesClass));
    }

    /**
     * Creates a ServiceLocator that will look up service classes in the provided Guice module.
     *
     * @param injector the injector to look in
     * @param serviceKeys the service classes that will be looked up
     */
    public GuiceServiceLocator(Injector injector, Set<ServiceKey> serviceKeys) {
        this.serviceKeys = serviceKeys;
        this.services = getServices(injector);
    }

    private Map<ServiceKey, Object> getServices(final Injector injector) {
        Map<ServiceKey, Object> services = Collections.synchronizedMap(new HashMap<ServiceKey, Object>());

        for (ServiceKey serviceKey : serviceKeys) {

            @SuppressWarnings("unchecked")
            Key<Object> key = serviceKey.getServiceName() != null
                    ? Key.<Object>get(serviceKey.getServiceType(), Names.named(serviceKey.getServiceName()))
                    : Key.<Object>get(serviceKey.getServiceType());

            services.put(serviceKey, injector.getInstance(key));
        }
        return Collections.unmodifiableMap(services);
    }

    public Set<ServiceKey> getServiceKeys() {
        return services.keySet();
    }

    public <T> T lookupService(ServiceKey<T> serviceKey) {
        return serviceKey.getServiceType().cast(services.get(serviceKey));
    }
}
